/*
    Sysepub 0.9.0 - A Symmetric Semi-Public key cipher
    Copyright (C) 2002 Pierre Blanc
    Pierre Blanc: blanc_teutoburgo@yahoo.it
    http://teutoburgo.cjb.net
    http://it.geocities.com/teutoburgo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    or go to      http://www.gnu.org/copyleft/gpl.html
*/
import java.io.*;
import java.util.*;
/**
 * Sysepub 0.9 - A Symmetric Semi-Public key cipher
 * Creation date: (30/08/2002 16.50.29)
 * @author: Pierre Blanc
 */
public class Sysepub {
    public static final int MAX_DIM_BLOCKS_FACTOR=2;
    public static String SSP_HOME = "";
    public static final String KEY_SMALL = "The public key is too small.. generate"
										+" a new file";
    public static final String GENERIC_ERROR = "Can't decrypt: try with earlier"
										+" seeds?";
	private int period=1, defaultLength=3000;
    private A1timeP a1t;
    Syxor s = new Syxor();
    Random rnd = new Random();
    private long seed = 0;
    private byte[] pattern;
    private String plain="plaintexts/", cipher="ciphertexts/",
	defaultFile="file", stringLog="";
    private String publicKey="publicKey/";
    private String patternString = "";
    private int patternLength = patternString.length();
    private boolean isGUI=false;
    /**
     * Sysepub constructor comment.
     */
    public Sysepub() {
    }

    public Sysepub(boolean isGUI) {
	super();
	SSP_HOME=new File((new File("")).getAbsolutePath()).getParent()+"/";
	String temp = "";
	Properties prop = new Properties();
	FileInputStream fis=null, fis2=null;
	try {
	    fis = new FileInputStream(new File(SSP_HOME+"config/ssp.cfg"));
		sspInit("ssp.cfg");
	} catch (FileNotFoundException fnfe){
		try{
			fis2 = new FileInputStream(new File("c:/ssp_home.cfg"));
		} catch (Exception e){
			sPrintLog("Error! Can't find Sysepub/config/ssp.cfg");
			sPrintLog("Can't find c:/ssp_home.cfg");
			sPrintLog("Under Windows copy it from Sysepub/patch4win/ssp_home.cfg");
			System.exit(-1);
		}
		try{
			prop.load(fis2);
			SSP_HOME=prop.getProperty("SSP_HOME");
            sspInit("ssp.cfg");
		} catch (FileNotFoundException e){
			e.printStackTrace();
			sPrintLog("Using SSP_HOME="+SSP_HOME);
			sPrintLog("Set the SSP_HOME");
			System.exit(-1);
		} catch(IOException ioe){
			ioe.printStackTrace();
		System.exit(-1);
		}
	} catch (IOException ioe) {
	    ioe.printStackTrace();
		System.exit(-1);
	} catch (NumberFormatException e){

	} catch (Exception e){
		e.printStackTrace();
	}
	this.isGUI=isGUI;
    }
    /**
	*	Used by the contructor and by a GUI method to load the properties
	*	in the configuration file
	*	@param fileName the name of the configuration file
	*/
	public void sspInit(String fileName) throws FileNotFoundException,
					IOException, NumberFormatException {
        //System.out.println("-----------reading "+fileName);
		String temp = "";
		Properties prop = new Properties();
		FileInputStream fis=null, fis2=null;

	    fis = new FileInputStream(new File(SSP_HOME+"config/"+fileName));
	    prop.load(fis);
	    temp = prop.getProperty("defaultfile","file");
	    defaultFile=temp;
	    temp = prop.getProperty("seed");
	    seed= Long.parseLong(temp);
	    patternString = prop.getProperty("patternstring");
        patternLength = patternString.length();
	    temp = prop.getProperty("period", "1");
	    period = Integer.parseInt(temp);
	    a1t = new A1timeP(period);
        temp = prop.getProperty("randomFileDefaultLength","3");
        defaultLength=Integer.parseInt(temp)*1024;
	}

	public void sCatchException (){
	  if (isGUI)
	  {
	  } else System.exit(-1);
	}
	/*
	* Catches an exception differently if it's within the command-line or
	* the GUI application.
	* @param message The exception message.
	*/
	public void sCatchException (String message){
	  if (isGUI)
	  {
		stringLog=stringLog+message+"\n";
	  } else System.exit(-1);
	}
	/*
	* Prints the log messages: in the GUI writes to a String, in the command-line
	* application to the system output.
	* @param s the message
	*/
	public void sPrintLog(String s){
      if (isGUI) stringLog=stringLog+s+"\n";
      else System.out.println(s);
    }
	/**
	* Returns to the GUI application the String of the log messages.
	*/
    public String flushLog(){
	  String output=stringLog+"\n";
	  stringLog="";
      return output;
    }   
	/**
     * Generates a random file in the publicKey directory
     * @param length the file length
     */
    public void generateRandomFile(int length){
      String fileName="publicKey";
      int fLength=fileName.length();
      int max=0;
      byte[] randomBytes = new byte[length];
      File f = new File(SSP_HOME+publicKey);
      File[] list = f.listFiles();
      try{
      for (int i=0; i<list.length; i++){
        if (!(fLength>list[i].getName().toString().length())){
          if(list[i].getName().toString().substring(0,fLength).equals(fileName))
          max++;
        }
      }
      }catch (Exception e){
        e.printStackTrace();
      }
      rnd.nextBytes(randomBytes);
      String progressive=Integer.toString(max+1);
      String generated=SSP_HOME+publicKey+fileName+progressive+".rnd";
      this.writeFile(generated,randomBytes);
      sPrintLog("Generated "+length+" random bytes in file "+generated);
    }
    /**
     * method for compacting the key files in a single array
     * @param mngs the array of the key files
     */
    public byte[] compactMngs(byte[][] mngs) {
		int i = 0, totalLength = 0, k = 0;
		for (i = 0; i < mngs.length; i++) {
			totalLength = totalLength + mngs[i].length;
		}
		byte[] compacted = new byte[totalLength];
		for (i = 0; i < mngs.length; i++) {
			for (int j = 0; j < mngs[i].length; j++) {
				compacted[k] = mngs[i][j];
				k++;
			}
		}
		return compacted;
    }
    /**
     * The method that performs decryption
     * @param x3 The ciphertext array
     * @param pubKey The pubKey array
     * @return The x plaintext array (x1 cleaned from random blocks b1, b2)
     */
    public byte[] decryptFast(byte[] x3, byte[] pubKey) throws Exception {
	int pubKeyL = pubKey.length;
	int l = x3.length;
	int i = 0, first = 0, last = 0, whereFound = 0;
	byte[] x1 = new byte[l];
	byte[] shifted = new byte[l];
	byte[] realKey = new byte[l];
	boolean notFound = false;
	byte[] x=null, prs;

    pattern = this.getPattern(patternString);
	sPrintLog("Decrypting x3 ...");
        prs=a1t.getBytes(l, seed);
	byte[] x2 = s.decrypt(x3, prs);
	sPrintLog("Decrypting x2 ...");
	int shift = a1t.getInt(seed, pubKeyL - l);
	//    sPrintLog("---> shift=" + shift + " <-------");
	for (i = 0; i < l; i++) {
	    realKey[i] = pubKey[i + shift];
	}
	x1 = s.decrypt(x2, realKey);
	whereFound = this.findPattern(pattern, x1, 0);
	if (whereFound != -1) {
	    //        sPrintLog("Pattern found");
	    first = whereFound;
	    whereFound = this.findPattern(pattern, x1, whereFound + 1);
	    if (whereFound != -1) {
		sPrintLog("Ok.");
		last = whereFound;
		shift = l + pubKeyL - 2 * patternLength;
	    }
	}
	//    sPrintLog("first " + first + " last " + last);
	if (last > 0) {
	    x = new byte[last - first - patternLength];
	    for (i = first + patternLength; i < last; i++) {
		x[i - first - patternLength] = x1[i];
	    }
	    sPrintLog("Decryption performed.");
	} else {
	    sPrintLog("Error - Can't decrypt");
	    //sCatchException();
		throw new Exception(GENERIC_ERROR);
	}
	return x;
    }
     /*
	 * Used to try decryption if first attempt failed.
     * @param x3 The ciphertext array
     * @param pubKey The pubKey array
	 * @param periods The number of tries
     * @return The x plaintext array (x1 cleaned from random blocks b1, b2)
	 */
    public byte[] tryDecryption(byte[] x3,
								byte[] pubKey, int periods) throws Exception {
	int pubKeyL = pubKey.length;
	int l = x3.length;
	int i = 0, first = 0, last = 0, whereFound = 0, k;
	byte[] x1 = new byte[l];
	byte[] shifted = new byte[l];
	byte[] realKey = new byte[l];
	boolean notFound = false;
	byte[] x=null, prs;

	if (pattern == null)
        pattern = this.getPattern(patternString);
	sPrintLog("Trying decryption..");
	for (k=1;k<periods+1 ;k++ ){
	  prs=a1t.getOtherBytes(l, seed, k);
  	  byte[] x2 = s.decrypt(x3, prs);
	  sPrintLog("Try number "+k);
	  int shift = a1t.getAnotherInt(seed, pubKeyL - l, k);
	  //    sPrintLog("---> shift=" + shift + " <-------");
	  for (i = 0; i < l; i++) {
	    realKey[i] = pubKey[i + shift];
	  }
	  x1 = s.decrypt(x2, realKey);
	  whereFound = this.findPattern(pattern, x1, 0);
	  if (whereFound != -1) {
	    //        sPrintLog("Pattern found");
	    first = whereFound;
	    whereFound = this.findPattern(pattern, x1, whereFound + 1);
	    if (whereFound != -1) {
		sPrintLog("Ok.");
		last = whereFound;
		shift = l + pubKeyL - 2 * patternLength;
	    }
	  }
	//    sPrintLog("first " + first + " last " + last);
	  if (last > 0) {
	      x = new byte[last - first - patternLength];
	      for (i = first + patternLength; i < last; i++) {
		    x[i - first - patternLength] = x1[i];
	      }
	      sPrintLog("Decryption performed.");
		  k=periods+2;
	  }
	}
	if (k==periods+1)
	{
		sPrintLog("Error - Can't decrypt after "+periods+" tries.");
	    sCatchException();
		throw new Exception(GENERIC_ERROR);
	}
	return x;
    }
    /**
     * Prepares for decryption and calls the decryptFast method <br>
	 * or the tryDecryption method
     * Writes the plaintext array x in its directory
     * @param pathName The pathName of the ciphertext
	 * @param mode The number of tries for tryDecryption. 0 to call decryptFast
     */
    public int doDecryption(String pathName, int mode) {
	byte[] x;
	byte[] x3;
    File f=new File(pathName);

	if(!f.isAbsolute()) 
		pathName=(new File(SSP_HOME+cipher+pathName)).getAbsolutePath();
	String fileName=f.getName();
	sPrintLog("Starting decryption of file "+pathName);

	try{
	byte[][] arrays = this.initArrays(pathName);
	x3 = arrays[0];
	byte[] pubKey = arrays[1];

	if (x3.length < pubKey.length)
	  if (mode==0)
	  	x = this.decryptFast(x3, pubKey);
	  else x = this.tryDecryption(x3, pubKey, mode);

	else {
	    sPrintLog("Error - The file is too big.. expand the"+
			       " publicKey files");
	    return -2;
	}
	this.writeFile(SSP_HOME + plain + fileName, x);
	sPrintLog("Written "+SSP_HOME + plain+fileName);
	if(patternLength<4) sPrintLog("WARNING!! The pattern used contains"+
			     " less than 4 characters..\nDecryption may have"+
			     " been performed with errors!");
	} catch (IOException e){
	  e.printStackTrace();
	  sCatchException(e.getMessage());
	} catch (Exception e){
	  sCatchException(e.getMessage());
	  return -1;
	}
	return 1;
	}
    /**
     * Prepares for encryption and calls the encryptFast method <br>
     * Writes the ciphertext array x3 in its directory
     * @param pathName The pathname of the plaintext
     */
    public void doEncryption(String pathName) {
	int i, j, l, pr = 0, maxShift = 0, minShift = 0, cLength = 0;
	byte[] x, x3;
    File f=new File(pathName);
	if(!f.isAbsolute()) 
		pathName=(new File(SSP_HOME+plain+pathName)).getAbsolutePath();
	String fileName=f.getName();
	sPrintLog("Starting encryption of file "+pathName);
	try{
	  byte[][] arrays = this.initArrays(pathName);
	  x = arrays[0];
	  byte[] pubKey = arrays[1];

	  x3 = this.encryptFast(x, pubKey);

	  if (x3.length > pubKey.length)
	      sPrintLog("Error - The file is too big.. expand the publicKey"+
			       " directory");
		  else{
			  this.writeFile(SSP_HOME + cipher+fileName, x3);
			  sPrintLog("Written "+SSP_HOME + cipher+fileName);
		  }
	  if(patternLength<4) sPrintLog("WARNING!! The pattern used contains"+
			     " less than 4 characters..\nDecryption may be"+
			     " performed with errors!");
	} catch (IOException e){
	  e.printStackTrace();
	  sCatchException(e.getMessage());
	} catch (Exception e){
	  sCatchException(e.getMessage());
	}
    }
    /**
     * The method that performs encryption
     * @param x The plaintext array
     * @param pubKey The pubKey array
     * @return The x3 ciphertext array
     */
    public byte[] encryptFast(byte[] x, byte[] pubKey) throws Exception {
	pattern = this.getPattern(patternString);
	patternLength=patternString.length();
	int l = x.length;
	int pubKeyL = pubKey.length;
	int lTot = l + 2 * this.patternLength;
	int factor=MAX_DIM_BLOCKS_FACTOR;
	int n1 = rnd.nextInt(factor * (l + 2 * patternLength));
	int n2 = rnd.nextInt(factor * (l + 2 * patternLength));
	byte[] b1 = new byte[n1];
	byte[] b2 = new byte[n2];
	int x1Length = l + n1 + n2 + 2 * patternLength;
	byte[] x1 = new byte[x1Length];
	byte[] x3, x2, prs;
	int i = 0, pXLength = 2 * patternLength + l;

	rnd.nextBytes(b1);
	rnd.nextBytes(b2);
	//*************************** appending the arrays ******
	for (i = 0; i < n1; i++) {
	    x1[i] = b1[i];
	}
	for (i = n1; i < n1 + patternLength; i++) {
	    x1[i] = pattern[i - n1];
	}
	for (i = n1 + patternLength; i < n1 + patternLength + l; i++) {
	    x1[i] = x[i - n1 - patternLength];
	}
	for (i = n1 + patternLength + l; i < n1 + pXLength; i++) {
	    x1[i] = pattern[i - (n1 + patternLength + l)];
	}
	for (i = n1 + pXLength; i < x1Length; i++) {
	    x1[i] = b2[i - (n1 + pXLength)];
	}
	//***********************************************************

	if (pubKeyL < x1Length) {
	    sPrintLog("Error - publicKey files are too small");
	    sCatchException();
		throw new Exception(KEY_SMALL);
	}
	byte[] realKey = new byte[x1Length];
	int maxShift = pubKeyL - x1Length;
	int pr = a1t.getInt(seed, maxShift);
	sPrintLog("Encrypting x2 ...");
	for (i = 0; i < x1Length; i++) {
	    realKey[i] = pubKey[i + pr];
	}
	x2 = s.encrypt(x1, realKey);
	sPrintLog("Encrypting x3 ...");
        prs=a1t.getBytes(x1Length, seed);
	x3 = s.encrypt(x2, prs);
	sPrintLog("Encryption performed.");
	return x3;
    }
    /**
     * Searches for the pattern within x1
     * @param pattern The pattern array
     * @param x1 The x1 array (x concatenated with the random blocks b1,b2)
     * @param start The starting index within x1
     * @return The index of the pattern in x1
     */
    public int findPattern(byte[] pattern, byte[] x1, int start) {
	boolean notFound = false;
	int whereFound = -1;
	for (int i = start; i < x1.length - pattern.length; i++) {
	    for (int j = 0; j < pattern.length; j++) {
		if (pattern[j] != x1[i + j]) {
		    notFound = true;
		    j = pattern.length;
		} else
		    notFound = false;
	    }
	    if (!notFound) {
		whereFound = i;
		i = x1.length - pattern.length;
	    }
	}
	return whereFound;
    }
    public String getDefaultFile(){
	return defaultFile;
    }
    public int getDefaultLength(){
	return defaultLength;
    }
/*    public byte[] getPattern(long seed) {
	String p = (new Long(seed)).toString();
	seedLength = p.length();
	byte[] pattern = new byte[seedLength];
	for (int ss = 0; ss < seedLength; ss++) {
	    pattern[ss] = (byte) p.charAt(ss);
	}
	return pattern;
    }*/
	public long getSeed(){
		return seed;
	}
	public void setSeed(long seed){
		this.seed=seed;
	}
	public String getPatternString(){
		return patternString;
	}
	public void setPatternString(String patternString){
		this.patternString=patternString;
	}
    /**
     * Creates the pattern array from a String
     * @param patternString The pattern string
     * @return The pattern array
     */
    public byte[] getPattern(String patternString) {
		int pLength = patternString.length();
		byte[] pattern = new byte[pLength];
		for (int ss = 0; ss < pLength; ss++) {
			pattern[ss] = (byte) patternString.charAt(ss);
		}
		return pattern;
    }
	public int getPubKeySize(){
		byte[][] mngs = this.readPubKey(SSP_HOME + publicKey);
		byte[] pubKey = this.compactMngs(mngs);
		return pubKey.length;
	}
    /**
     * Makes the initialization of the arrays of plaintext/ciphertext and pubKey
     * @param pathname The pathName of the plaintext/ciphertext
     * @return Two arrays: fileToXOR the file to be XORed and pubKey the
     * public key
     */
    public byte[][] initArrays(String pathName) throws IOException {
	byte[][] arrays=new byte[2][];
	byte[] fileToXOR=null;

   	fileToXOR = this.readFile(pathName);
	byte[][] mngs = this.readPubKey(SSP_HOME + publicKey);
	byte[] pubKey = this.compactMngs(mngs);
	arrays[0]=fileToXOR;
	arrays[1]=pubKey;

	return arrays;
    }

	public static void main(String args[]) {
	Sysepub ssp = new Sysepub(false);
	PrintMessages pm = new PrintMessages();
        pm.printCopyright();
	int argsLen = args.length;
	String fileName = ssp.getDefaultFile();
        int length=ssp.getDefaultLength();
	String badParameter="Bad parameter.. type: java Sysepub -h";

	if (argsLen == 0) {
	    pm.printHelp();
	} else
	    if (argsLen == 1) {
		if (args[0].equalsIgnoreCase("e"))
		    ssp.doEncryption(fileName);
		else
		    if (args[0].equalsIgnoreCase("d"))
			ssp.doDecryption(fileName, 0);
		    else
			if (args[0].equalsIgnoreCase("h")
			    || args[0].equalsIgnoreCase("help")
			    || args[0].equalsIgnoreCase("-h")
			    || args[0].equalsIgnoreCase("--help"))
			    pm.printHelp();
                        else if (args[0].equalsIgnoreCase("g"))
                          ssp.generateRandomFile(length);
			else if (args[0].equalsIgnoreCase("w")){
                          pm.printNoWarranty();
                        }else
			    System.out.println(badParameter);
	    } else
		if (argsLen == 2)
		    if (args[0].equalsIgnoreCase("e"))
			ssp.doEncryption(args[1]);
		    else
			if (args[0].equalsIgnoreCase("d"))
			    ssp.doDecryption(args[1], 0);
                        else if (args[0].equalsIgnoreCase("g")){
                          try{
                            length=Integer.parseInt(args[1])*1024;
                          } catch (NumberFormatException nfe){
                            nfe.printStackTrace();
                            System.exit(-1);
                          }
                          ssp.generateRandomFile(length);
                        }
			else
			    System.out.println(badParameter);
		else if (argsLen == 3)
			if (args[0].equalsIgnoreCase("t"))
			{
				int periods=3;
				try{
					periods=Integer.parseInt(args[2]);
				} catch (NumberFormatException nfe){
                            nfe.printStackTrace();
                            System.exit(-1);
				}
				ssp.doDecryption(args[1],periods);
			}else
			    System.out.println(badParameter);
		else
		    System.out.println("Wrong number of parameters "+argsLen);
    }

    public byte[] readFile(String pathName) throws IOException {
	File f = new File(pathName);
	long fLength = f.length();
	byte[] file = new byte[(int) fLength];

	    FileInputStream fis = new FileInputStream(f);
	    fis.read(file);

	return file;
    }
    /**
     * Reads the array pubKey from the publicKey directory
     * @param pathName The pathname
     * @return The array pubKey
     */
    public byte[][] readPubKey(String pathName) {
	File f = new File(pathName);
	File[] list = f.listFiles();
	int len = list.length, i = 0, j = 0;
	long fLength = 0;
	byte[][] mngs = new byte[len][];
	sPrintLog("Reading "+len+" public key files:");
	for (i = 0; i < len; i++) {
	    sPrintLog(list[i].toString());
	    if (list[i].isFile()) {
		fLength = list[i].length();
		try {
		    mngs[i] = new byte[(int) fLength];
		    (new FileInputStream(list[i])).read(mngs[i]);
		} catch (IOException ioe) {
		    ioe.printStackTrace();
                    sCatchException();
		}
		j++;
	    }
	}

	byte[][] pubKey = new byte[j][];
		for (i = 0; i < j; i++) {
			pubKey[i] = mngs[i];
		}
		return pubKey;
    }

	public void writeFile(String name, byte[] bytes) {
	try {
	    FileOutputStream fos = new FileOutputStream(name);
	    fos.write(bytes);
	} catch (FileNotFoundException fnfe) {
	    fnfe.printStackTrace();
            sCatchException();
	} catch (IOException ioe){
	    ioe.printStackTrace();
            sCatchException();
	}
    }
}
